# -*- coding: utf-8 -*-

import time
import logging
import random
import functools
import threading
from boolparser import BoolParser

class GpioEmu(object):
    HIGH = 1
    LOW = 0
    IN = 10
    OUT = 11
    BOARD = 20
    BCM = 21
    RISING = 30
    FALLING = 31
    BOTH = 32
    
    @classmethod
    def setmode(cls, mode):
        print("setmode {}".format(mode))

    @classmethod
    def setup(cls,port, mode):
        print("setup {} {}".format(port, mode))

    @classmethod
    def input(cls, port):
        print("input {}".format(port))
        return cls.HIGH

    @classmethod
    def output(cls, port, state):
        print("output {} {}".format(port, state))

    @classmethod
    def cleanup(cls, channels=None):
        print("cleanup {}".format(channels))

    @classmethod
    def wait_for_edge(cls, channel, edge, timeout=1000):
        print("wait_for_edge {} {} {}".format(channel, edge, timeout))
        ret = random.choice([True, False, None])
        if ret is None:
            time.sleep(timeout/1000.0)
        return ret

    @classmethod
    def add_event_detect(cls, channel, edge, callback=None, bouncetime=None):
        print("add_event_detect {} {} {} {}".format(channel, edge, callback, bouncetime))

    @classmethod
    def event_detected(cls, channel):
        print ("event_dectected {}".format(channel))
        return random.choice([True, False])

    @classmethod
    def remove_event_detect(cls, channel):
        print("remove_event_detect {}".format(channel))

try:
    from RPi import GPIO as _gpio
except ImportError:
    _gpio = GpioEmu()

   

class RelaisStatus(object):

    def __init__(self, config):
        self.config = config  # AntConfig
        self.pttstatus = {} # trxid to ptt (Bool)
        self.pttporttotrx = {}
        self.switchstatus = {} # switchid to switchstatus (Bool)
        self.currentstatus = {} # relais id to  relaystatus (bool)
        self.trx_frequency = {} #trxid to frequency in Hz
        self.relais_parser = {} # rid to boolparser
        self.symbols = [] # list of tuples symbols in config (name, Boolparser)
        self.relais_list = []
        self._run = False
        self._gpio_lock = threading.Lock()
        self._init_from_config()
        self._inverted_relais = True

    def frequency_to_band(self, f):
        """
        :para f: Frequncy in Hz
        """
        _bands = [("2190m" , 0.1357, 0.1378),
                  ("630m" , 0.472, 0.479),
                  ("560m", 0.501, 0.504),
                  ("160m", 1.8, 2.0),
                  ("80m", 3.5 , 4.0),
                  ("60m", 5.06, 5.45),
                  ("40m", 7.0, 7.3),
                  ("30m", 10.1, 10.15),
                  ("20m", 14.0, 14.35),
                  ("17m", 18.068, 18.168),
                  ("15m", 21.0, 21.450),
                  ("12m", 24.890, 24.99),
                  ("10m", 28.0, 29.7),
                  ("6m", 50.0, 54.0),
                  ("4m", 70.0, 71.0),
                  ("2m", 144.0, 148.0),
                  ("70cm", 420.0, 450.0),
                  ("23cm", 1240.0, 1300.0),
                  ("13cm",  2300.0, 2450.0),
                  ("6cm", 5650.0, 5925.0),
                  ("3cm", 10000.0, 10500.0),
                  ("1.25cm", 24000.0, 24250),
                  ("6mm", 47000.0, 47200.0)]
        band = "UNKOWN"
        for b, l, u in _bands:
            if f >= l*1e6 and f <= u*1e6:
                band = b
                break
        return band
    
    def _to_rel(self, rel_val):
        return self._inverted_relais ^ bool(rel_val)

    def _init_from_config(self):
        """
        inits state from config
        """
        self._inverted_relais = self.config.globalconf.get("inverted_relais")
        for swid, swconf in self.config.switches.items():
            self.switch_enabled(swid, swconf.default)
        _gpio.setmode(_gpio.BCM)
        for sym in self.config.symbols:
            self.symbols.append((sym.name, BoolParser(sym.expression)))
        rids = sorted(list(self.config.relais))
        for rid in rids:
            relais = self.config.relais[rid]
            self.relais_list.append(relais)
            channel = relais.gpioport
            _gpio.setup(channel, _gpio.OUT)
            _gpio.output(channel, self._to_rel(relais.default))
            self.currentstatus[rid] = relais.default
            self.relais_parser[rid] = BoolParser(relais.expression)
            
        for trxid, trxc in self.config.transceivers.items():
            self.pttstatus[trxid] = False
            self.trx_frequency[trxid] = 0.0
            _gpio.setup(trxc.pttgpio, _gpio.IN)
            ptt_callback = functools.partial(self.ptt_toggle, trxc.pttgpio)
            _gpio.add_event_detect(trxc.pttgpio,
                                   _gpio.BOTH,
                                   callback=ptt_callback,
                                   bouncetime=10)
            self.pttporttotrx[trxc.pttgpio] = trxid

    def ptt_toggle(self, *args):
        """
        :return relaisid to relay status (bool) 
        """
        gpioport = args[0]
        logging.info("ptt_input toggle on gpio port %s", gpioport)
        value = not _gpio.input(gpioport) # we got neg ptt
        self.pttstatus[self.pttporttotrx[gpioport]] = value
        self.set_relais()

    def switch_enabled(self, switchid, value):
        """
        :return dict relaisid -> to relay status (bool)
        """
        logging.debug("switch by id id: %s value:%s", switchid, value)
        self.switchstatus[switchid] = value
        # think abaout rturnung values
        self.set_relais()
        return self.relais_status()

    def switch_toggle_by_number(self, number):
        """
        toggle switch by number

        :return relais status

        """
        logging.info("toggle switch by number: %s", number)
        switch = self.config.switch_list[number]
        if switch is not None:
            return self.switch_toggle_by_id(switch.id)
        else:
            # return current status
            return self.relais_status()

    def switch_toggle_by_id(self, switchid):
        """
        :return relais status
        """
        logging.info("toggle switch by id: %s", switchid)
        value = self.switchstatus.get(switchid, False)
        self.switchstatus[switchid] = not value
        # think abaout rturnung values
        self.set_relais()
        return self.relais_status()

    def set_qrg(self, trxid, frequency):
        """
        sets the qrg for trx
        """
        logging.info("set trx frequency trx:%s freq:%s", trxid, frequency)
        self.trx_frequency[trxid] = frequency
        self.set_relais()

    def _evaluate(self):
        """
        :return status for all relais  dict relaisid -> to relay status (bool)
        """
        variables = {}
        for trx, f in self.trx_frequency.items():
            variables["f_{}".format(trx)] = f
            b = self.frequency_to_band(f)
            logging.debug("Band to %s for trx: %s", b, trx) 
            variables["band_{}".format(trx)] = b
        for trx, ptt in self.pttstatus.items():
            variables["PTT_{}".format(trx)] = ptt
        for sw, swstatus in self.switchstatus.items():
            variables[sw] = swstatus
        for sname, sParser in self.symbols:
            valid, needed_variables = sParser.get_variables()
            missing_vars = needed_variables - set(variables.keys())
            # must be checked in advance
            if missing_vars:
                logging.warning("warning: missing vars for symbol:%s %s", sname, missing_vars)
            if not valid:
                logging.warning("warning: invalid vars for symbol:%s %s", sname, missing_vars)
            else:
                variables[sname] = sParser.eval(variables)
        relais_status = {}
        logging.debug("evaluate: variables: %s", variables)

        for rid, b_parser in self.relais_parser.items():
            valid, needed_variables = b_parser.get_variables()
            missing_vars = needed_variables - set(variables.keys())
            # must be checked in advance
            if missing_vars:
                logging.warning("warning: missing vars: %s", missing_vars)
            if not valid:
                logging.warning("warning: invalid vars: %s", missing_vars)
            r_value = b_parser.eval(variables)
            relais_status[rid] = r_value
            logging.debug("relais: %s, expression: '%s' = %s", rid, b_parser.expression, r_value)

        return relais_status
    
    def _calc_diff(self, new_values):
        """
        :return only relais to toggle. This are two lists
            list 1: relais to switch off
            list 2: relais to swith on
        """
        off_relais = []
        on_relais = []
        r_ids = sorted(self.currentstatus.keys())
        for r_id in r_ids:
            c_val = self.currentstatus.get(r_id)
            n_val = new_values.get(r_id)
            if n_val is not None and c_val != n_val:
                if n_val:
                    on_relais.append(r_id)
                else:
                    off_relais.append(r_id)
        return off_relais, on_relais

    def toggle_relais(self, relais_off, relais_on):
        """
        toggles gpio relais
        and remember last values in self.currentstatus
        """
        for r_id in relais_off:
            relais = self.config.relais[r_id]
            if relais.delayoff > 0:
                time.sleep(relais.delayoff)
            _gpio.output(relais.gpioport, self._to_rel(False))
        for r_id in relais_on:
            relais = self.config.relais[r_id]
            if relais.delayon > 0:
                time.sleep(relais.delayon)
            _gpio.output(relais.gpioport, self._to_rel(True))

    def set_relais(self):
        """
        evaulate current relais status
        and sets relais
        """
        self._gpio_lock.acquire()
        try:
            relais_status = self._evaluate()
            new_off, new_on = self._calc_diff(relais_status)
            self.toggle_relais(new_off, new_on)
            self.currentstatus = relais_status
        finally:
            self._gpio_lock.release()

    def _execute(self):
        """
        sets relais from current knwon status
        must me called to toggle all relais
        """
        while self._run:
            self.set_relais()
            # must be fast (tx delay)
            time.sleep(0.05)
    
    def run(self):
        self._run = True
        t =  threading.Thread(target=self._execute)
        t.start()

    def close(self):
        self._run = False
        for trxc in self.config.transceivers.values():
            _gpio.remove_event_detect(trxc.pttgpio)
        _gpio.cleanup()

    def relais_status(self):
        rel_status = {}
        for r_id, state in self.currentstatus.items():
            rel_status["rel_{}".format(r_id)] = state
        trx_f = {}
        for t_id, f in self.trx_frequency.items():
            trx_f[t_id] = "{:10.3f} Mhz".format(f/1.e6)
        return {"switches": self.switchstatus,
                "relais": rel_status,
                "frequencies": trx_f}

