#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Antenna / accesssory switch remote control

(c) dg7bbp 2021-2022
"""
import config_reader
import argparse
# import wsgiserver
import logging
import waitress
import time
import antennastatus
import threading
import sys
import os
from more.chameleon import ChameleonApp
from static_resources import ResourcesApp, get_fileresource
from hamlib import HamLibConnection
from fileutil import get_install_dir
_verbose = 0

_version = "0.9e"

def show_error(msg):
    sys.stderr.write("{}\n".format(msg))

_verbose_level = 0

def log(level, msg):
    global _verbose_level
    if level <= _verbose_level:
        print(msg)


def analyse_args():
    ap = argparse.ArgumentParser(description="7accctrl version: {}".format(_version))
    ap.add_argument("--port", type=int, help="Web service listen port default: 8000)")
    ap.add_argument("--hostbind", help="Web service local bind address default:localhost")
    ap.add_argument("--config", help="configfile for relais configuration")
    ap.add_argument("--verbose", type=int, default=0, choices=[0,1,2,3,4,5], help="print some logging 5 = max")
    args = ap.parse_args()
    return args


class RigBind(object):
    def __init__(self, dst_hosts, r_status):
        """
        :param dst_host: list of Transceiver (config) object
        :param r_status: Relaisstatus
        """
        self.dst_hosts = dst_hosts
        self.r_status = r_status
        self.run = False
        self._hamlib_threads = []
        for trx in self.dst_hosts:
            self._hamlib_threads.append(threading.Thread(target=self.hamlib_handler,
                                                        args=(trx,)))

    def start(self):
        self.run=True
        for t in self._hamlib_threads:
            t.start()

    def stop(self):
        self.run=False

    def hamlib_handler(self, trx):
        last_frequency = 0.0
        while self.run:
            try:
                with HamLibConnection(trx.rigctld) as trx_conn:
                    while self.run:
                        f = trx_conn.read_frequency()
                        if f is None:
                            self.r_status.set_qrg(trx.id, 0.0)
                            last_frequency = 0.0
                            raise ValueError("Connection may failed frequency is None")
                        elif f != last_frequency:
                            self.r_status.set_qrg(trx.id, f)
                            last_frequency = f
                        time.sleep(0.5)
            except Exception as e:
                show_error("ConnectionError: %s\n"
                           "gqrx or hamlib not running." % e)
                        
_relais_status = None
_rig_bind = None


class RootApp(ChameleonApp):
    pass

class SwModel(object):
    def __init__(self, cmd, switch):
        self.cmd = cmd
        self.switch = switch

    def execute_cmd(self):
        global _relais_status
        if self.cmd == "status":
            return _relais_status.relais_status()
        elif self.cmd == "swon":
            return _relais_status.switch_enabled(self.switch, True)
        elif self.cmd == "swoff":
            _relais_status.switch_enabled(self.switch, False)
            return _relais_status.relais_status()
        elif self.cmd == "swtoggle":
            _relais_status.switch_toggle_by_number(int(self.switch))
        else:
            return {"error": "unknown command"}
            
@RootApp.path(path='')
class Root(object):
    pass

@RootApp.template_directory()
def get_panel_template_dir():
    t_dir = os.path.join(get_install_dir(),
                         "template") 
    return t_dir

@RootApp.path(model=SwModel, path='7ant/{cmd}/{switch}' )
def run_sw_cmd(cmd, switch):
    return SwModel(cmd, switch)


@RootApp.html(model=Root, template="index.pt")
def main_page(self, _request):
    # returns config for panel
    # switches
    # relais
    # transceivers
    switches = []
    for cnt, sw in enumerate(_relais_status.config.switch_list):
        sw_dict = sw.to_dict()
        sw_dict["number"] = cnt
        sw_dict["url"] = "switchToggle({});".format(cnt)
        switches.append(sw_dict)
    trx_list = [trx.to_dict() for trx in _relais_status.config.transceivers.values()]
    relais = [r.to_dict() for r in _relais_status.relais_list]
    config = {"switches": switches,
              "transceiver": trx_list, 
              "relais": relais}
    return config
    #with open(os.path.join(os.path.dirname(__file__), "web", "index.html"), "rb") as f:
    #    html = f.read()
    #return html

@RootApp.json(model=SwModel) 
def antsw_cmd(self, request):
    return self.execute_cmd()

@RootApp.mount(app=ResourcesApp, path="web")
def mount_static_resources():
    return ResourcesApp()

def run_wsgi_server():
    global _relais_status
    global _rig_bind
    args = analyse_args()
    if args.verbose > 0:
        loglevel = 60-args.verbose*10
        logging.basicConfig(stream=sys.stdout,level=loglevel)
        logging.info("Starting 7accctrl version: %s",  _version)
    config_file = args.config
    if config_file is None:
        config_file = os.path.join(os.environ["HOME"],
                                   ".config",
                                   "7accctrl",
                                   "7accctrl.config")
    if not os.path.isfile(config_file):
        show_error("Config file {} not found".format(config_file))
        return -1
    r_config = config_reader.AntConfig(config_file)
    errors = r_config.verify()
    if not errors:
        _relais_status = antennastatus.RelaisStatus(r_config)
        tranceivers = list(r_config.transceivers.values())
        _rig_bind = RigBind(tranceivers, _relais_status)
        _rig_bind.start()
        host = r_config.globalconf.get("bindhost")
        port = r_config.globalconf.get("bindport")
        if not args.hostbind:
            if host is None:
                host = "localhost"
        else:
            host = args.hostbind
        if not args.port:
            if port is None:
                port = 8000
        else:
            port = args.port
        #  only relais are set direct after set_operation
        # _relais_status.run()
        try:
            #server = wsgiserver.WSGIServer(RootApp(),
            #                               port=args.port,
            #                               host=args.hostbind)
            #server.start()
            waitress.serve(RootApp(),
                            host=host,
                            port=port)
        finally:
            # stop threads
            _rig_bind.stop()
            _relais_status.close()
    else:
        for e in errors:
            print(e)
    return 0


