#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Antenna / accesssory switch remote control

(c) dg7bbp 2021-2022


config reader for antenna switch

configs are json structures. We support up to 8 relais outputs (GPIO
and gui switches. <switchids> can be used in relais output expressions. 
Additional input values are PTT_<trxid> and f_<trxid> for drequency in Hz
and band_<trxid> for band descriiption like 20m, 70cm, ...
see adif definitions.

Symbols are variables. In symbol n+1 symbol n can be used.
                       In relais expression all symbols can
                       be used. reserved names are not allowed
                       but are not checked

Config format:
Relais start with 0
{ "global": {"inverted_relais" : true,
             "bindhost": "localhost",
             "bindport": 8000
            },
 "symbols": [{"name": "<symbolname>",
              "expression: "<symbolexpression>"}]
 "switches": [ {"id": <switch id>,
                "name": "<visible name>",
                "default": <bool>}},
             ]
 "relais" : {"<relais number (int)>": {"name": "<visible name for status in GUI>,
                               "expression": "Python expression">,
                               "gpioport": <bcm port optional if not standard>",
                               "delayon": 0,
                               "delayoff": 0}, 
            },
"transceiver": {"trxid": {"name": "<visible name>",
                          "rigctld": "<address of rigctld for this trx",
                          "pttgpio": <GPIO-Inputport for PTT (active low)>",
"switchgroups": [{"caption": "<name of group">,
                  "switches": ["<switchid>",..]
                   
                             
Example expressions: 
   enable PA PTT_<trxid> for 2m PA if enabled via SW0
   f_trxid1 >= 144e6 and  f_trxid1 < 146e6 ) and PTT_trxid1 and SW0
   
   antenna switch use  short wave beam
   f_trxid1 >14e6 and f_trxid1 < 30e6 
   
   antenna switch use v/uhf vertical(SW1 switches between vertical and beam)
   we don't have a hoizontalantenna for 70cm. Use an 70 cm always
   f_trxid1> 144e6 and f_trxid1 <440e6 and (not SW1 or f_trxid1 > 430e6)
   
   use 2m horziontal
   f_trxid1> 144e6 and f_trxid1 <146e6 and SW1
   
   splitted rx antenna for 80m enabled by SW2
   
   f_trxid1 >3.5e6 and f_trxid1 <3.8e6 and not PTT_trxid1 and SW2

   use 23cm transverter on SW3
   
   f_trxid1 >144e6 and f_trxid1 <148e6 and SW3
   
   enabled 23cm PA PTT_trxid1 with SW 4
   
   f_trxid1 >144e6 and f_trxid1 <148e6 and SW3 and SW4 and PTT_trxid1

   Relais are switched off in descendig order and switched on
   in ascendig order. The PTT for PA should have always the
   highest number.
   If possible a TRX should be configured with a tx delay.

"""

import json
from boolparser import BoolParser

class Switch(object):
    def __init__(self, switchid, values):
        self.id = switchid
        self.name = values.get("name", self.id)
        self.default = values.get("default", False)

    def to_dict(self):
        return {"id": self.id,
                "name": self.name}


class Relais(object):
    def_ports = {0: 5,
                 1: 6,
                 2: 13, 
                 3: 16,
                 4: 19,
                 5: 20,
                 6: 21,
                 7: 26}

    def __init__(self, relid, values):
        self.id = relid
        self.name = values.get("name", "R{}".format(self.id))
        self.expression = values.get("expression", "False")
        self.delayon = values.get("delayon", 0)
        self.delayoff = values.get("delayoff", 0)
        self.gpioport = values.get("gpioport", self.def_ports[relid])
        self.default = values.get("default", False)

    def to_dict(self):
        return {"id": "rel_{}".format(self.id),
                "name": self.name}


class Transceiver(object):
    def __init__(self, trxid, values):
        self.id = trxid
        self.name = values.get("name")
        self.rigctld = values.get("rigctld")
        self.pttgpio = values.get("pttgpio")

    def to_dict(self):
            return {"id": self.id,
                    "name": self.name}    

class Symbol(object):
    def __init__(self, jsondata):
        self.name = jsondata.get("name")
        self.expression = jsondata.get("expression")
        if not self.name:
            raise ValueError("Symbolname not defined")
        if not self.expression:
            raise ValueError("Symbolexpression not defined") 

class AntConfig(object):
    
    
    def __init__(self, fname=None):
        self.switches = dict()  # Switch ID to Switch
        self.switch_list = []   # list of Switches
        self.relais = dict() # Relais number to  Relais
        self.transceivers = dict() # trx to Transceiver
        self.globalconf = dict() 
        self.symbols = [] # list of symbol
        if fname is not None:
            self.read(fname)
 
    def read(self, filename):
        with open(filename) as f:
            data = json.load(f)
            self.globalconf = data.get("global", 
                                       {"inverted_relais": True})
            for swItem in data.get("symbols", []):
                self.symbols.append(Symbol(swItem))
            switches = data.get("switches")
            for sw_values in switches:
                sw_id = sw_values["id"]
                switch = Switch(sw_id, sw_values)
                self.switch_list.append(switch)
                self.switches[sw_id] = switch
            relais = data.get("relais")
            for r_id_str, r_values in relais.items():
                r_id = int(r_id_str)
                self.relais[int(r_id)]  = Relais(r_id, r_values)
            trx_dict = data.get("transceiver")
            for trx_id, trx_value in trx_dict.items():
                self.transceivers[trx_id] = Transceiver(trx_id, trx_value)
                
                
    def verify(self):
        """
        verify expressions of relais
        returns list of failed relais
        """
        errors = []
        valid_vars = set()
        for trx_id in self.transceivers.keys():
            valid_vars.add("f_{}".format(trx_id))
            valid_vars.add("band_{}".format(trx_id))
            valid_vars.add("PTT_{}".format(trx_id))        
        valid_vars.update(self.switches.keys())
        for s in self.symbols:
            b = BoolParser(s.expression)
            ok, used_vars= b.get_variables()
            if not ok:
                errors.append("Symbol '{}' has invalid syntax '{}'".format(s.name, s.expression))
            unknown_vars  = used_vars - valid_vars
            if unknown_vars:
                errors.append("Symbol '{}' expression is using unkown variables {}".format(s.name, unknown_vars))
            valid_vars.add(s.name)
        for r in self.relais.values():
            b = BoolParser(r.expression)
            ok, used_vars= b.get_variables()
            if not ok:
                errors.append("Relais '{}' has invalid syntax '{}'".format(r.id, r.expression))
            unknown_vars  = used_vars - valid_vars
            if unknown_vars:
                errors.append("Relais '{}' expression is using unkown variables {}".format(r.id,
                                                                                           ", ".join(list(unknown_vars))))
        return errors   
